from typing import Any
from dataclasses import asdict

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table

from .data import DataModel
from .data import GridColumn
from .data import DataInfo
from . import column
from . import layout_table
from .column import CellItem


def build(proj: Project, map: Mapping, dm: DataModel, table: Table, data: Any) -> DataInfo:
    items = column.get_axis_items(proj, map, dm.settings, dm.x_axis, dm.instance)
    cells = column.make_cell_item(proj, map, dm, table)
    columns = _build_columns(dm, items, cells)

    row_items = column.get_axis_items(proj, map, dm.settings, dm.y_axis, dm.instance)
    matrix = {k['value']: {dm.y_axis: k['value'], '_sort_key': k['caption']} for k in row_items}
    for d in data:
        y_data = d[dm.y_axis]
        if y_data not in matrix:
            continue
        for cell in cells:
            matrix[y_data][f"{d[dm.x_axis]}_{cell.column_name}"] = d[cell.column_name]
    rows = [column.build_one_row(columns, d) for d in sorted(matrix.values(), key=lambda x: x['_sort_key'])]

    return DataInfo(
        grid_columns=columns,
        grid_rows=rows,
        allow_line_addition_and_removal=False
    )


def _build_columns(dm: DataModel, items: list[object], cells: list[CellItem]) -> list[GridColumn]:
    columns = []
    columns.append(column.make_grid_column(
        dm.y_axis,
        '',
        editable=False
    ))
    columns.extend([
        column.make_grid_column(
            **asdict(cell, dict_factory=column.exclude_names),
            column_name=f"{item['value']}_{cell.column_name}",
            display_name=f"{item['caption']}({cell.display_name})"
        ) for cell in cells for item in items
    ])
    return columns


def parse(proj: Project, map: Mapping, dm: DataModel, table: Table, rows: object) -> list[dict[str, Any]]:
    items = column.get_axis_items(proj, map, dm.settings, dm.x_axis, dm.instance)
    expdata = [
        {
            dm.y_axis: row[dm.y_axis],
            dm.x_axis: col['value'],
            **{cell: row[f"{col['value']}_{cell}"] for cell in dm.cells}
        }
        for col in items for row in rows
    ]
    return layout_table.parse(proj, map, dm, table, expdata)



# ファイル名
FNAME_MAPPING = '_mapping.yaml'

# レイアウト
LAYOUT_TABLE = 'table'
LAYOUT_MATRIX = 'matrix'
LAYOUT_SINGLE = 'single'

# カラム定義
BIND_TYPE_FIXED = 'fixed'
BIND_TYPE_SELECTABLE = 'selectable'
BIND_TYPE_CALL = 'call'
# カラム定義（特殊タイプ）
BIND_TYPE_REF_ID = 'ref_id'
BIND_TYPE_FOREIGN_KEY = "foreign_key"
BIND_TYPE_EMBEDDED_DATA = "embedded_data"

# callのタイプ
CALL_TYPE_UUID = 'uuid'

# FIXME セルタイプ
FIELD_TYPE_STRING = 'string'
FIELD_TYPE_SELECTABLE = 'singleSelect'
FIELD_TYPE_SYSTEM = 'system'

# 同期モード
SYNC_MODE_DROP_CREATE = 'drop_create'



from typing import Any
from dataclasses import dataclass
from uuid import uuid4

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table
from ..schema import find_field
from ..fileio import load_data
from ..fileio import load_model
from ..fileio import get_data_model_name
from .. import const

from .data import DataModel
from .data import GridColumn


def make_grid_column(
        column_name: str,
        display_name: str,
        *,
        type: str = const.FIELD_TYPE_STRING,
        width: int = 150,
        editable: bool = True,
        hide: bool = False,
        items: list | None = None,
        fixed_value: str | None = None,
        call_value: str | None = None,
        reference: str | None = None):
    '''
    GridColumnを生成する。
    '''
    return GridColumn(
        field=column_name,
        type=type,
        header_name=display_name,
        width=width,
        editable=editable,
        hide=hide,
        items=items,
        fixed_value=fixed_value,
        call_value=call_value,
        reference=reference
    )


def load_for_select_items(folder: str, map: Mapping, ins: str, ref: str):
    '''
    マスタデータをロードして、選択肢用データを生成する。
    '''
    items = load_data(folder, map.id, ins, ref, True)
    if items is None:
        if map.parent is not None:
            items = load_for_select_items(folder, map.parent, ins, ref)
    else:
        # データが見つかった場合には、DataModelをロードし、
        # caption, valueのデータを生成する
        dm: DataModel = load_model(
            get_data_model_name(folder, map.id, ins, ref),
            DataModel,
            id=map.id,
            instance=ins,
            table_name=ref
        )
        if dm.caption is not None and dm.value is not None:
            items = [{
                'caption': item[dm.caption],
                'value': item[dm.value]
            } for item in items]
        else:
            raise RuntimeError('Invalid Data Model for Master Data.')
    return items


@dataclass
class CellItem:
    column_name: str
    display_name: str
    type: str = const.FIELD_TYPE_STRING
    editable: bool = True
    items: list[object] | None = None
    fixed_value: str | None = None
    call_value: str | None = None
    reference: str | None = None


def make_cell_item(proj: Project, map: Mapping, dm: DataModel, table: Table) -> list[CellItem]:
    '''
    セルの設定を生成する。
    '''
    result = []
    for cell in dm.cells:
        field = find_field(table.fields, cell)
        if cell in dm.settings:
            setting = dm.settings[cell]
            if setting['type'] == const.BIND_TYPE_FOREIGN_KEY:
                items = load_for_select_items(proj.folder, map, dm.instance, setting['value'])

                result.append(CellItem(
                    column_name=field.column_name,
                    display_name=field.display_name,
                    type=const.FIELD_TYPE_SELECTABLE,
                    editable=True,
                    items=items
                ))
            elif setting['type'] == const.BIND_TYPE_EMBEDDED_DATA:
                result.append(CellItem(
                    column_name=field.column_name,
                    display_name=field.display_name,
                    type=const.FIELD_TYPE_SELECTABLE,
                    editable=True,
                    items=setting['values']
                ))
            elif setting['type'] == const.BIND_TYPE_REF_ID:
                result.append(CellItem(
                    column_name=field.column_name,
                    display_name=field.display_name,
                    type=const.FIELD_TYPE_STRING,
                    editable=False,
                    reference='id'
                ))
            else:
                bind = proj.bindings[setting['type']]
                if bind.type == const.BIND_TYPE_FIXED:
                    result.append(CellItem(
                        column_name=field.column_name,
                        display_name=field.display_name,
                        editable=False,
                        fixed_value=bind.value
                    ))
                elif bind.type == const.BIND_TYPE_CALL:
                    result.append(CellItem(
                        column_name=field.column_name,
                        display_name=field.display_name,
                        editable=False,
                        call_value=bind.value
                    ))
                elif bind.type == const.BIND_TYPE_SELECTABLE:
                    result.append(CellItem(
                        column_name=field.column_name,
                        display_name=field.display_name,
                        items=bind.items
                    ))
        else:
            result.append(CellItem(
                column_name=field.column_name,
                display_name=field.display_name
            ))
    return result


def exclude_names(items: list[tuple[str, Any]]) -> dict[str, Any]:
    '''
    column_name, display_nameを除外した辞書を生成する。
    '''
    return {key: value for key, value in items if key not in ['column_name', 'display_name']}


def adjust_column_value(col: GridColumn, value: Any) -> Any:
    '''
    valueの中にフィールドのキーがなかったら、値を補完する。
    '''
    if col.field in value:
        return value[col.field]
    if col.fixed_value is not None:
        return col.fixed_value
    if col.call_value is not None:
        if col.call_value == const.CALL_TYPE_UUID:
            return str(uuid4())
    return ''


def build_one_row(columns: list[GridColumn], data: Any, need_id: bool = True) -> dict[str, Any]:
    '''
    1行分のデータを生成する。
    '''
    row = {
        col.field: adjust_column_value(col, data)
        for col in columns
    }
    if need_id and 'id' not in row:
        row['id'] = str(uuid4())
    return row


def get_axis_items(proj: Project, map: Mapping, settings: dict[str, object], axis: str, ins: str) -> list[object]:
    items = None
    setting = settings[axis]
    if setting['type'] == const.BIND_TYPE_FOREIGN_KEY:
        items = load_for_select_items(proj.folder, map, ins, setting['value'])
    elif setting['type'] == const.BIND_TYPE_EMBEDDED_DATA:
        items = setting['values']
    elif setting['type'] in proj.bindings:
        bind = proj.bindings[setting['type']]
        if bind.type == const.BIND_TYPE_SELECTABLE:
            items = bind.items
    if items is None:
        raise RuntimeError(f'No Data for axis. ({ins}@{axis})')
    return items



from typing import Any
from uuid import uuid4

from ..project import Project
from ..environ.data import Mapping
from ..schema import Table
from ..schema import Field
from .. import const

from .data import DataModel
from .data import GridColumn
from .data import DataInfo
from . import column


def build(proj: Project, map: Mapping, dm: DataModel, table: Table, data: Any) -> DataInfo:
    # フィールドを設定に従って展開する。ForeignKeyなども設定に含まれるものとする。
    columns = _build_columns(proj, map, dm, table)
    rows = [column.build_one_row(columns, d) for d in data]

    return DataInfo(
        grid_columns=columns,
        grid_rows=rows,
        allow_line_addition_and_removal=True
    )


def build_one_row(proj: Project, map: Mapping, dm: DataModel, table: Table) -> dict[str, Any]:
    columns = _build_columns(proj, map, dm, table)
    return column.build_one_row(columns, {})


def _build_columns(proj: Project, map: Mapping, dm: DataModel, table: Table) -> list[GridColumn]:
    columns = []
    for field in table.fields:
        if field.column_name in dm.settings:
            grid_column = _make_grid_column_from_setting(proj, map, dm, field)
        else:
            grid_column = column.make_grid_column(
                field.column_name,
                field.display_name)

        columns.append(grid_column)
    return columns


def _make_grid_column_from_setting(proj: Project, map: Mapping, dm: DataModel, field: Field):
    setting = dm.settings[field.column_name]
    if setting['type'] == const.BIND_TYPE_REF_ID:
        # 固定値と同じものとする
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            editable=False,
            hide=True,
            reference='id')
    if setting['type'] == const.BIND_TYPE_FOREIGN_KEY:
        # 選択肢は外部データを参照する以外は選択型と同じ
        items = column.load_for_select_items(proj.folder, map, dm.instance, setting['value'])
        if items is None:
            # データが見つからなければ空の配列とする
            items = []
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            items=items)
    if setting['type'] == const.BIND_TYPE_EMBEDDED_DATA:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            items=setting['values'])

    bind = proj.bindings[setting['type']]
    if bind.type == const.BIND_TYPE_FIXED:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            editable=False,
            hide=True,
            fixed_value=bind.value)
    if bind.type == const.BIND_TYPE_CALL:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            editable=False,
            call_value=bind.value)
    if bind.type == const.BIND_TYPE_SELECTABLE:
        return column.make_grid_column(
            field.column_name,
            field.display_name,
            items=bind.items)
    raise RuntimeError('Unknown Data Type')


def parse(proj: Project, map: Mapping, dm: DataModel, table: Table, rows: object) -> list[dict[str, Any]]:
    # idカラムを除いて返却する。
    columns = _build_columns(proj, map, dm, table)
    data = [column.build_one_row(columns, d, False) for d in rows]
    return data



from logging import getLogger
from .dbio import engine
from .dbio import database
from .dbio import table

from .models.environ import mapping
from .models.environ.data import Mapping
from .models.environ import entity
from .models.fileio import load_data

logger = getLogger(__name__)


def _load_for_entry(folder: str, map: Mapping, ins: str, tbl: str):
    items = load_data(folder, map.id, ins, tbl, True)
    if items is None:
        if map.parent is not None:
            items = _load_for_entry(folder, map.parent, ins, tbl)
    return items


def apply(project, env, deploy):
    map = mapping.get(project.folder, env)
    with engine.get_connection(project.deployments[deploy]) as conn:
        # データベース作成
        logger.info(f'database {map.id}')
        if database.is_exist(conn, map.id):
            database.drop(conn, map.id)
        database.create(conn, map.id)
        # database.create_user(conn, 'test', '%', 'test')
        # database.grant(conn, env, 'test', '%')

        # テーブル作成
        for ins in map.instances:
            schema = project.schemas[ins]
            for tbl in schema.tables.values():
                logger.info(f'table {map.id}.{tbl.table_name}')
                table.create(conn, map.id, tbl)

        # データ投入
        for ins in map.instances:
            schema = project.schemas[ins]
            for tbl in schema.tables.values():
                items = _load_for_entry(project.folder, map, ins, tbl.table_name)
                if items is not None:
                    logger.info(f'insert {map.id}.{tbl.table_name}')
                    table.insert(conn, map.id, tbl, items)



from fastapi import APIRouter
from fastapi import Request
from fastapi import Body

from ..models.project import project
from ..models.environ import mapping
from ..models.environ import entity

from .dtos import Result
from .dtos import Data
from .dtos import NewMapping
from .dtos import NewDataModel
from .dtos import convert_to_data_filename
from .dtos import convert_to_mapping
from .dtos import convert_to_data_model

router = APIRouter(prefix='/environs')


@router.get('')
def get_mappings(request: Request):
    proj = project(request)
    maps = mapping.items(proj.folder)
    return Result(data=maps)


@router.post('/{id}')
def create_mapping(id: str, data: NewMapping, request: Request) -> Result:
    proj = project(request)
    if mapping.is_exist(proj.folder, id):
        return Result(
            status='ERROR',
            message='ERROR_EXIST_MAPPING'
        )
    mapping.save(proj.folder, id, convert_to_mapping(id, data))
    return Result()


@router.get('/{id}/tables')
def get_tables(id: str, request: Request):
    '''
    既に存在するテーブル一覧。
    '''
    proj = project(request)
    tables = [
        convert_to_data_filename(ins, tbl)
        for ins, tbl in entity.items(proj.schemas, proj.folder, id)
    ]
    return Result(data=tables)


@router.get('/{id}/init')
def get_not_exist_tables(id: str, request: Request):
    '''
    新規作成時のテーブル一覧。既に存在するものは除く
    '''
    proj = project(request)
    tables = [
        convert_to_data_filename(ins, tbl)
        for ins, tbl in entity.items(proj.schemas, proj.folder, id, exist=False)
    ]
    return Result(data=tables)


@router.get('/{id}/tables/{instance}/{table}')
def get_table(id: str, instance: str, table: str, request: Request):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    dm, tbl, info = entity.get(proj, map, instance, table)
    data = Data(
        model=dm,
        info=info,
        table=tbl
    )
    return Result(data=data)


@router.get('/{id}/tables/{instance}/{table}/row')
def get_row(id: str, instance: str, table: str, request: Request):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    row = entity.get_row(proj, map, instance, table)
    return Result(data=row)


@router.post('/{id}/tables/{instance}/{table}')
def create_data_model(id: str, instance: str, table: str, data: NewDataModel, request: Request):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    entity.save(proj, map, instance, table, convert_to_data_model(data))
    return Result()


@router.put('/{id}/tables/{instance}/{table}')
def update_data(id: str, instance: str, table: str, request: Request, body=Body(...)):
    proj = project(request)
    map = mapping.get(proj.folder, id)
    entity.save_data(proj, map, instance, table, body)
    return Result()
